VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsFaceDetection"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
' Autor: F. Schler (frank@activevb.de)
' Datum: 06/2022

Option Explicit

' Basiert auf:
' https://www.autohotkey.com/boards/viewtopic.php?t=72797

' ----==== Const ====----
Private Const S_OK As Long = &H0&
Private Const CC_STDCALL As Long = &H4&
Private Const GdiPlusVersion As Long = &H1&

Private Const WindowsGraphicsImagingBitmapDecoder As String = "Windows.Graphics.Imaging.BitmapDecoder"
Private Const WindowsGraphicsImagingSoftwareBitmap As String = "Windows.Graphics.Imaging.SoftwareBitmap"
Private Const WindowsMediaFaceAnalysisFaceDetector As String = "Windows.Media.FaceAnalysis.FaceDetector"

Private Const IID_IClosable As String = "{30d5a829-7fa4-4026-83bb-d75bae4ea99e}"
Private Const IID_IAsyncInfo As String = "{00000036-0000-0000-c000-000000000046}"
Private Const IID_IBitmapFrame As String = "{72a49a1c-8081-438d-91bc-94ecfc8185c6}"
Private Const IID_IRandomAccessStream As String = "{905a0fe1-bc53-11df-8c49-001e4fc686da}"
Private Const IID_IFaceDetectorStatics As String = "{bc042d67-9047-33f6-881b-6746c1b218b8}"
Private Const IID_IBitmapDecoderStatics As String = "{438ccb26-bcef-4e95-bad6-23a822e58d01}"
Private Const IID_ISoftwareBitmapStatics As String = "{df0385db-672f-4a9d-806e-c2442f343e86}"
Private Const IID_IBitmapFrameWithSoftwareBitmap As String = "{fe287c9a-420c-4963-87ad-691436e08383}"

Private Const CLSID_GDIPPngEncoder As String = "{557cf406-1a04-11d3-9a73-0000f81ef32e}"

' ----==== Enums ====----
Private Enum vtb_Interfaces
    
    ' IUnknown
    IUnknown_QueryInterface = 0
    IUnknown_Release = 2

    ' IInspectable
    IInspectable_GetIids = 3
    IInspectable_GetRuntimeClassName = 4
    IInspectable_GetTrustLevel = 5

    'IFaceDetectorStatics
    IFaceDetectorStatics_CreateAsync = 6
    IFaceDetectorStatics_GetSupportedBitmapPixelFormats = 7

    'IReadOnlyList
    IReadOnlyList_GetAt = 6
    IReadOnlyList_GetSize = 7
    
    ' IBitmapDecoderStatics
    IBitmapDecoderStatics_CreateAsync = 14
    
    ' IBitmapFrame
    IBitmapFrame_GetBitmapPixelFormat = 8

    ' IBitmapFrameWithSoftwareBitmap
    IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync = 6
    
    ' ISoftwareBitmapStatics
    ISoftwareBitmapStatics_Convert = 7
    
    ' IFaceDetector
    IFaceDetector_DetectFacesAsync = 6
    
    ' IDetectedFace
    IDetectedFace_GetFaceBox = 6
    
    ' IAsyncInfo
    IAsyncInfo_GetStatus = 7
    IAsyncInfo_Close = 10

    ' IAsyncInterface
    IAsyncInterface_GetResults = 8
    
    ' IClosable
    IClosable_Close = 6

End Enum

Private Enum FileAccessMode
    Read = 0
    ReadWrite = 1
End Enum

Private Enum BSOS_OPTIONS
    BSOS_DEFAULT = 0
    BSOS_PREFERDESTINATIONSTREAM = 1
End Enum

Private Enum AsyncStatus
    Started = 0
    Completed = 1
    Canceled = 2
    Error = 3
End Enum

Private Enum TrustLevel
    BaseTrust = 0
    PartialTrust = 1
    FullTrust = 2
End Enum

Public Enum BitmapPixelFormat
    BitmapPixelFormat_Unknown = 0
    BitmapPixelFormat_Rgba16 = 12
    BitmapPixelFormat_Rgba8 = 30
    BitmapPixelFormat_Gray16 = 57
    BitmapPixelFormat_Gray8 = 62
    BitmapPixelFormat_Bgra8 = 87
    BitmapPixelFormat_Nv12 = 103
    BitmapPixelFormat_Yuy2 = 107
End Enum

Private Enum Status
    OK = 0
    GenericError = 1
    InvalidParameter = 2
    OutOfMemory = 3
    ObjectBusy = 4
    InsufficientBuffer = 5
    NotImplemented = 6
    Win32Error = 7
    WrongState = 8
    Aborted = 9
    FileNotFound = 10
    ValueOverflow = 11
    AccessDenied = 12
    UnknownImageFormat = 13
    FontFamilyNotFound = 14
    FontStyleNotFound = 15
    NotTrueTypeFont = 16
    UnsupportedGdiplusVersion = 17
    GdiplusNotInitialized = 18
    PropertyNotFound = 19
    PropertyNotSupported = 20
    ProfileNotFound = 21
End Enum

' ----==== Types ====----
Private Type GUID
    data1 As Long
    data2 As Long
    data3 As Long
    data4 As Long
End Type

Private Type GDIPlusStartupInput
    GdiPlusVersion As Long
    DebugEventCallback As Long
    SuppressBackgroundThread As Long
    SuppressExternalCodecs As Long
End Type

Private Type GdiplusStartupOutput
    NotificationHook As Long
    NotificationUnhook As Long
End Type

' ----==== Ole32.dll Declarations ====----
Private Declare Function CLSIDFromString Lib "Ole32.dll" ( _
                         ByVal pstring As Long, _
                         ByRef pCLSID As GUID) As Long
                         
Private Declare Function StringFromGUID Lib "Ole32.dll" _
                         Alias "StringFromGUID2" ( _
                         ByRef rguid As GUID, _
                         ByVal lpsz As Long, _
                         ByVal cchMax As Long) As Long
                         
Private Declare Sub CoTaskMemFree Lib "Ole32.dll" ( _
                    ByVal hMem As Long)
                         
' ----==== Oleaut32.dll Declarations ====----
Private Declare Function DispCallFunc Lib "Oleaut32.dll" ( _
                         ByVal pvInstance As Long, _
                         ByVal oVft As Long, _
                         ByVal cc As Long, _
                         ByVal vtReturn As VbVarType, _
                         ByVal cActuals As Long, _
                         ByVal prgvt As Any, _
                         ByVal prgpvarg As Any, _
                         ByRef pvargResult As Variant) As Long

' ----==== Kernel32.dll Declarations ====----
Private Declare Sub CopyMemory Lib "Kernel32.dll" _
                    Alias "RtlMoveMemory" ( _
                    ByRef pTo As Any, _
                    ByRef uFrom As Any, _
                    ByVal lSize As Long)

Private Declare Function lstrlenW Lib "Kernel32.dll" ( _
                         ByVal lpString As Long) As Long

' ----==== Combase.dll Declarations ====----
Private Declare Function WindowsCreateString Lib "Combase.dll" ( _
                         ByVal sourceString As Long, _
                         ByVal lenght As Long, _
                         ByRef hString As Long) As Long

Private Declare Function WindowsDeleteString Lib "Combase.dll" ( _
                         ByVal sourceString As Long) As Long

Private Declare Function RoGetActivationFactory Lib "Combase.dll" ( _
                         ByVal activatableClassId As Long, _
                         ByRef riid As GUID, _
                         ByRef factory As Long) As Long

' ----==== Shlwapi.dll Declarations ====----
Private Declare Function SHCreateMemStream Lib "Shlwapi.dll" ( _
                         ByVal pInit As Long, _
                         ByVal cbInit As Long) As Long

' ----==== ShCore.dll Declarations ====----
Private Declare Function CreateRandomAccessStreamOverStream Lib "ShCore.dll" ( _
                         ByVal pStream As Long, _
                         ByVal options As BSOS_OPTIONS, _
                         ByRef riid As GUID, _
                         ByRef ppv As Long) As Long

' ----==== Gdiplus.dll Declarations ====----
Private Declare Function GdiplusShutdown Lib "GdiPlus.dll" ( _
                         ByVal token As Long) As Status
                         
Private Declare Function GdiplusStartup Lib "GdiPlus.dll" ( _
                         ByRef token As Long, _
                         ByRef lpInput As GDIPlusStartupInput, _
                         ByRef lpOutput As GdiplusStartupOutput) As Status

Private Declare Function GdipCreateBitmapFromHBITMAP Lib "GdiPlus.dll" ( _
                         ByVal hbm As Long, _
                         ByVal hpal As Long, _
                         ByRef bitmap As Long) As Status

Private Declare Function GdipDisposeImage Lib "GdiPlus.dll" ( _
                         ByVal image As Long) As Status

Private Declare Function GdipSaveImageToStream Lib "GdiPlus.dll" ( _
                         ByVal mImage As Long, _
                         ByVal mStream As Long, _
                         ByRef mClsidEncoder As GUID, _
                         ByVal mEncoderParams As Long) As Status

' ----==== Variablen ====----
Private m_lngGdipToken As Long
Private m_bolIsFaceDetectorInitialized As Boolean
Private m_lngDetectedFacesCount As Long
Private m_lngBitmapPixelFormatCount As Long
Private m_tDetectedFacesBounds() As BitmapBounds
Private m_eBitmapPixelFormats() As BitmapPixelFormat
Private m_pIFaceDetector As Long
Private m_pIBitmapDecoderStatics As Long
Private m_pISoftwareBitmapStatics As Long

' ----==== Class ====----
Private Sub Class_Initialize()
    m_bolIsFaceDetectorInitialized = False
    m_lngDetectedFacesCount = 0&
    m_lngBitmapPixelFormatCount = 0&
    m_pIFaceDetector = 0&
    m_pIBitmapDecoderStatics = 0&
    m_pISoftwareBitmapStatics = 0&
    
    If GetActivationFactory(WindowsGraphicsImagingBitmapDecoder, _
                            IID_IBitmapDecoderStatics, _
                            m_pIBitmapDecoderStatics) Then
        If GetActivationFactory(WindowsGraphicsImagingSoftwareBitmap, _
                                IID_ISoftwareBitmapStatics, _
                                m_pISoftwareBitmapStatics) Then
            Dim pIFaceDetectorStatics As Long
            If GetActivationFactory(WindowsMediaFaceAnalysisFaceDetector, _
                                    IID_IFaceDetectorStatics, _
                                    pIFaceDetectorStatics) Then
                If Invoke(pIFaceDetectorStatics, _
                            IFaceDetectorStatics_CreateAsync, _
                            VarPtr(m_pIFaceDetector)) = S_OK Then
                    If WaitForAsync(m_pIFaceDetector) Then
                        Dim tGdipStartupInput As GDIPlusStartupInput
                        Dim tGdipStartupOutput As GdiplusStartupOutput
                        tGdipStartupInput.GdiPlusVersion = GdiPlusVersion
                        If GdiplusStartup(m_lngGdipToken, tGdipStartupInput, _
                                          tGdipStartupOutput) = OK Then
                            m_bolIsFaceDetectorInitialized = True
                        End If
                    End If
                End If
                Call GetSupportedBitmapPixelFormats(pIFaceDetectorStatics)
                Call Release(pIFaceDetectorStatics)
            End If
        End If
    End If
End Sub

Private Sub Class_Terminate()
    Call Release(m_pIFaceDetector)
    Call Release(m_pISoftwareBitmapStatics)
    Call Release(m_pIBitmapDecoderStatics)
    If m_lngGdipToken <> 0& Then Call GdiplusShutdown(m_lngGdipToken)
End Sub

' ----==== Properties ====----
Public Property Get GetIsFaceDetectorInitialized() As Boolean
    GetIsFaceDetectorInitialized = m_bolIsFaceDetectorInitialized
End Property

Public Property Get GetSupportedBitmapPixelFormatCount() As Long
    GetSupportedBitmapPixelFormatCount = m_lngBitmapPixelFormatCount
End Property

Public Property Get GetSupportedBitmapPixelFormatByIndex(ByVal lngIndex As Long) As BitmapPixelFormat
    If lngIndex >= 0 And lngIndex < Me.GetSupportedBitmapPixelFormatCount Then
        GetSupportedBitmapPixelFormatByIndex = m_eBitmapPixelFormats(lngIndex)
    End If
End Property

Public Property Get GetSupportedBitmapPixelFormatByIndexAsString(ByVal lngIndex As Long) As String
    If lngIndex >= 0 And lngIndex < Me.GetSupportedBitmapPixelFormatCount Then
        Dim strPixelFormat As String
        Select Case m_eBitmapPixelFormats(lngIndex)
            Case BitmapPixelFormat.BitmapPixelFormat_Unknown
                strPixelFormat = "BitmapPixelFormat_Unknown"
            Case BitmapPixelFormat.BitmapPixelFormat_Rgba16
                strPixelFormat = "BitmapPixelFormat_Rgba16"
            Case BitmapPixelFormat.BitmapPixelFormat_Rgba8
                strPixelFormat = "BitmapPixelFormat_Rgba8"
            Case BitmapPixelFormat.BitmapPixelFormat_Gray16
                strPixelFormat = "BitmapPixelFormat_Gray16"
            Case BitmapPixelFormat.BitmapPixelFormat_Gray8
                strPixelFormat = "BitmapPixelFormat_Gray8"
            Case BitmapPixelFormat.BitmapPixelFormat_Bgra8
                strPixelFormat = "BitmapPixelFormat_Bgra8"
            Case BitmapPixelFormat.BitmapPixelFormat_Nv12
                strPixelFormat = "BitmapPixelFormat_Nv12"
            Case BitmapPixelFormat.BitmapPixelFormat_Yuy2
                strPixelFormat = "BitmapPixelFormat_Yuy2"
        End Select
        GetSupportedBitmapPixelFormatByIndexAsString = strPixelFormat
    End If
End Property

Public Property Get GetDetectedFacesCount() As Long
    GetDetectedFacesCount = m_lngDetectedFacesCount
End Property

Friend Property Get GetDetectedFaceBitmapBoundsByIndex(ByVal lngIndex As Long) As BitmapBounds
    If lngIndex >= 0 And lngIndex < Me.GetDetectedFacesCount Then
        GetDetectedFaceBitmapBoundsByIndex = m_tDetectedFacesBounds(lngIndex)
    End If
End Property

' ----==== Subs & Functions ====----
Public Function GetFaceDetectionFromFromHBitmap(ByVal hBitmap As Long, _
                                                ByVal eSupportedPixelFormat As BitmapPixelFormat) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    If m_bolIsFaceDetectorInitialized And hBitmap <> 0& Then
        Dim pBitmap As Long
        If GdipCreateBitmapFromHBITMAP(hBitmap, 0&, pBitmap) = OK Then
            Dim pIStream As Long
            pIStream = SHCreateMemStream(0&, 0&)
            If pIStream <> 0& Then
                If GdipSaveImageToStream(pBitmap, pIStream, _
                                         Str2Guid(CLSID_GDIPPngEncoder), 0&) = OK Then
                    Dim pIRandomAccessStream As Long
                    If CreateRandomAccessStreamOverStream(pIStream, BSOS_DEFAULT, _
                                                          Str2Guid(IID_IRandomAccessStream), _
                                                          pIRandomAccessStream) = S_OK Then
                        bolRet = GetFaceDetectionFromIRandomAccessStream(pIRandomAccessStream, _
                                                                         eSupportedPixelFormat)
                        Call CloseAndRelease(pIRandomAccessStream)
                    End If
                End If
                Call Release(pIStream)
            End If
            Call GdipDisposeImage(pBitmap)
        End If
    End If
    GetFaceDetectionFromFromHBitmap = bolRet
End Function

Private Function GetFaceDetectionFromIRandomAccessStream(ByVal pIRandomAccessStream As Long, _
                                                         ByVal eSupportedPixelFormat As BitmapPixelFormat) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    Dim pISoftwareBitmap As Long
    If GetISoftwareBitmapFromIRandomAccessStream(pIRandomAccessStream, _
                                                 eSupportedPixelFormat, _
                                                 pISoftwareBitmap) Then
        Dim pIReadOnlyList As Long
        If Invoke(m_pIFaceDetector, IFaceDetector_DetectFacesAsync, _
                  pISoftwareBitmap, VarPtr(pIReadOnlyList)) = S_OK Then
            If WaitForAsync(pIReadOnlyList) Then
                If Invoke(pIReadOnlyList, IReadOnlyList_GetSize, _
                          VarPtr(m_lngDetectedFacesCount)) = S_OK Then
                    If m_lngDetectedFacesCount > 0 Then
                        ReDim m_tDetectedFacesBounds(m_lngDetectedFacesCount - 1)
                        Dim lngIndex As Long
                        For lngIndex = 0 To m_lngDetectedFacesCount - 1
                            Dim pIDetectedFace As Long
                            If Invoke(pIReadOnlyList, IReadOnlyList_GetAt, _
                                      lngIndex, VarPtr(pIDetectedFace)) = S_OK Then
                                Dim tBitmapBounds As BitmapBounds
                                If Invoke(pIDetectedFace, _
                                          IDetectedFace_GetFaceBox, _
                                          VarPtr(tBitmapBounds)) = S_OK Then
                                    m_tDetectedFacesBounds(lngIndex) = tBitmapBounds
                                End If
                                Call Release(pIDetectedFace)
                            End If
                        Next
                        bolRet = True
                    End If
                End If
            End If
            Call Release(pIReadOnlyList)
        End If
        Call CloseAndRelease(pISoftwareBitmap)
    End If
    GetFaceDetectionFromIRandomAccessStream = bolRet
End Function

Private Function GetISoftwareBitmapFromIRandomAccessStream(ByVal pIRandomAccessStream As Long, _
                                                           ByVal eSupportedPixelFormat As BitmapPixelFormat, _
                                                           ByRef pISoftwareBitmap As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    If m_bolIsFaceDetectorInitialized Then
        Dim pIBitmapDecoder As Long
        If Invoke(m_pIBitmapDecoderStatics, IBitmapDecoderStatics_CreateAsync, _
                  pIRandomAccessStream, VarPtr(pIBitmapDecoder)) = S_OK Then
            If WaitForAsync(pIBitmapDecoder) Then
                Dim pIBitmapFrame As Long
                If Invoke(pIBitmapDecoder, IUnknown_QueryInterface, _
                          VarPtr(Str2Guid(IID_IBitmapFrame)), VarPtr(pIBitmapFrame)) = S_OK Then
                    Dim eBitmapPixelFormat As BitmapPixelFormat
                    Call Invoke(pIBitmapFrame, IBitmapFrame_GetBitmapPixelFormat, VarPtr(eBitmapPixelFormat))
                    Dim lngCount As Long
                    lngCount = Me.GetSupportedBitmapPixelFormatCount
                    If lngCount > 0 Then
                        Dim lngIndex As Long
                        Dim bolIsBitmapPixelFormatSupported As Boolean
                        bolIsBitmapPixelFormatSupported = False
                        For lngIndex = 0 To lngCount - 1
                            If eBitmapPixelFormat = Me.GetSupportedBitmapPixelFormatByIndex(lngIndex) Then
                                bolIsBitmapPixelFormatSupported = True
                                Exit For
                            End If
                        Next
                        Dim pIBitmapFrameWithSoftwareBitmap As Long
                        If Invoke(pIBitmapDecoder, IUnknown_QueryInterface, _
                                  VarPtr(Str2Guid(IID_IBitmapFrameWithSoftwareBitmap)), _
                                  VarPtr(pIBitmapFrameWithSoftwareBitmap)) = S_OK Then
                            If Invoke(pIBitmapFrameWithSoftwareBitmap, _
                                      IBitmapFrameWithSoftwareBitmap_GetSoftwareBitmapAsync, _
                                      VarPtr(pISoftwareBitmap)) = S_OK Then
                                If WaitForAsync(pISoftwareBitmap) Then
                                    bolRet = True
                                End If
                            End If
                            Call Release(pIBitmapFrameWithSoftwareBitmap)
                        End If
                        If Not bolIsBitmapPixelFormatSupported Then
                            bolRet = False
                            Dim pISoftwareBitmapTmp As Long
                            If Invoke(m_pISoftwareBitmapStatics, _
                                      ISoftwareBitmapStatics_Convert, _
                                      pISoftwareBitmap, eSupportedPixelFormat, _
                                      VarPtr(pISoftwareBitmapTmp)) = S_OK Then
                                Call CloseAndRelease(pISoftwareBitmap)
                                pISoftwareBitmap = pISoftwareBitmapTmp
                                bolRet = True
                            End If
                        End If
                    End If
                    Call Release(pIBitmapFrame)
                End If
            End If
            Call Release(pIBitmapDecoder)
        End If
    End If
    GetISoftwareBitmapFromIRandomAccessStream = bolRet
End Function

Private Function GetSupportedBitmapPixelFormats(ByVal pIFaceDetectorStatics As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    Dim pIReadOnlyList As Long
    If Invoke(pIFaceDetectorStatics, _
              IFaceDetectorStatics_GetSupportedBitmapPixelFormats, _
              VarPtr(pIReadOnlyList)) = S_OK Then
        If Invoke(pIReadOnlyList, _
                  IReadOnlyList_GetSize, _
                  VarPtr(m_lngBitmapPixelFormatCount)) = S_OK Then
            If m_lngBitmapPixelFormatCount > 0 Then
                ReDim m_eBitmapPixelFormats(m_lngBitmapPixelFormatCount - 1)
                Dim lngIndex As Long
                For lngIndex = 0 To m_lngBitmapPixelFormatCount - 1
                    Dim eBitmapPixelFormat As BitmapPixelFormat
                    If Invoke(pIReadOnlyList, IReadOnlyList_GetAt, lngIndex, _
                        VarPtr(eBitmapPixelFormat)) = S_OK Then
                        m_eBitmapPixelFormats(lngIndex) = eBitmapPixelFormat
                    End If
                Next
                bolRet = True
            End If
        End If
        Call Release(pIReadOnlyList)
    End If
    GetSupportedBitmapPixelFormats = bolRet
End Function

Private Function GetActivationFactory(ByVal ClassName As String, _
                                      ByVal IID As String, _
                                      ByRef pFactory As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    Dim hString As Long
    If WindowsCreateString(StrPtr(ClassName), Len(ClassName), hString) = S_OK Then
        If hString <> 0& Then
            If RoGetActivationFactory(hString, Str2Guid(IID), pFactory) = S_OK Then bolRet = True
            Call WindowsDeleteString(hString)
        End If
    End If
    GetActivationFactory = bolRet
End Function

Private Function WaitForAsync(ByRef pInterface As Long) As Boolean
    Dim bolRet As Boolean
    bolRet = False
    If pInterface <> 0& Then
        Dim pIAsyncInfo As Long
        If Invoke(pInterface, IUnknown_QueryInterface, _
                  VarPtr(Str2Guid(IID_IAsyncInfo)), _
                  VarPtr(pIAsyncInfo)) = S_OK Then
            Dim eStatus As AsyncStatus
            Do
                If Invoke(pIAsyncInfo, IAsyncInfo_GetStatus, _
                          VarPtr(eStatus)) = S_OK Then DoEvents
            Loop While eStatus = Started
            If eStatus = Completed Then
                bolRet = True
            Else
                Call Invoke(pIAsyncInfo, IAsyncInfo_Close)
            End If
            Call Release(pIAsyncInfo)
            If bolRet Then
                Dim pRet As Long
                If Invoke(pInterface, IAsyncInterface_GetResults, VarPtr(pRet)) = S_OK Then
                    Call Release(pInterface)
                    pInterface = pRet
                End If
            Else
                Call Release(pInterface)
            End If
        End If
    End If
    WaitForAsync = bolRet
End Function

Private Sub Release(ByRef pInterface As Long)
    Dim lngRet As Long
    If pInterface <> 0& Then
        Call Invoke(pInterface, IUnknown_Release)
        pInterface = 0&
    End If
End Sub

Private Function CloseAndRelease(ByRef pInterface As Long) As Long
    Dim lngRet As Long
    If pInterface <> 0& Then
        Dim pIClosable As Long
        If Invoke(pInterface, IUnknown_QueryInterface, _
                  VarPtr(Str2Guid(IID_IClosable)), _
                  VarPtr(pIClosable)) = S_OK Then
            Call Invoke(pIClosable, IClosable_Close)
            Call Release(pIClosable)
            Call Release(pInterface)
            pInterface = 0&
        End If
    End If
    CloseAndRelease = lngRet
End Function

Private Sub GetInspectableInfo(ByVal pInspectableInterface As Long, _
                               ByVal strInterface As String)
    If pInspectableInterface <> 0& Then
        Debug.Print "Inspectable Info for: " & strInterface
        Debug.Print "RuntimeClassName = " & GetInspectable_GetRuntimeClassName(pInspectableInterface)
        Dim eTrustLevel As TrustLevel
        eTrustLevel = GetInspectable_GetTrustLevel(pInspectableInterface)
        Select Case eTrustLevel
            Case TrustLevel.BaseTrust
                Debug.Print "TrustLevel = BaseTrust"
            Case TrustLevel.FullTrust
                Debug.Print "TrustLevel = FullTrust"
            Case TrustLevel.PartialTrust
                Debug.Print "TrustLevel = PartialTrust"
            Case Else
                Debug.Print "TrustLevel = UnknownTrust"
        End Select
        Dim tIIDs() As GUID
        Dim lngIIDsCount As Long
        lngIIDsCount = GetInspectable_GetIIDs(pInspectableInterface, tIIDs)
        Debug.Print "AviableIIDsCount = " & CStr(lngIIDsCount)
        If lngIIDsCount > 0 Then
            Dim lngIIDsItem As Long
            For lngIIDsItem = 0 To lngIIDsCount - 1
                Debug.Print "    AviableIID " & CStr(lngIIDsItem + 1) & _
                            " = " & Guid2Str(tIIDs(lngIIDsItem))
            Next
        End If
        Debug.Print String(60, "-")
    End If
End Sub

Private Function GetInspectable_GetIIDs(ByVal pInterface As Long, _
                                        ByRef tIIDs() As GUID) As Long
    Dim lngIIDCount As Long
    If pInterface <> 0& Then
        Dim pIIDs As Long
        If Invoke(pInterface, IInspectable_GetIids, _
                  VarPtr(lngIIDCount), VarPtr(pIIDs)) = S_OK Then
            If lngIIDCount > 0 Then
                Dim lngBytes As Long
                Dim lngIIDItem As Long
                ReDim tIIDs(lngIIDCount - 1)
                For lngIIDItem = 0 To lngIIDCount - 1
                    lngBytes = Len(tIIDs(lngIIDItem))
                    Call CopyMemory(tIIDs(lngIIDItem), _
                                    ByVal pIIDs + (lngBytes * lngIIDItem), _
                                    lngBytes)
                Next
            End If
        End If
    End If
    GetInspectable_GetIIDs = lngIIDCount
End Function

Private Function GetInspectable_GetRuntimeClassName(ByVal pInterface As Long) As String
    Dim strRet As String
    If pInterface <> 0& Then
        Dim pStr As Long
        If Invoke(pInterface, IInspectable_GetRuntimeClassName, _
                  VarPtr(pStr)) = S_OK Then
            If pStr <> 0& Then strRet = Ptr2Str(pStr)
        End If
    End If
    GetInspectable_GetRuntimeClassName = strRet
End Function

Private Function GetInspectable_GetTrustLevel(ByVal pInterface As Long) As TrustLevel
    Dim eRet As TrustLevel
    If pInterface <> 0& Then
        Dim lngValue As Long
        If Invoke(pInterface, IInspectable_GetTrustLevel, _
                  VarPtr(lngValue)) = S_OK Then
            eRet = lngValue
        End If
    End If
    GetInspectable_GetTrustLevel = eRet
End Function


Private Function Invoke(ByVal pInterface As Long, _
                        ByVal vtb As vtb_Interfaces, _
                        ParamArray var()) As Variant
    If pInterface <> 0& Then
        Invoke = OleInvoke(pInterface, vtb, var)
    End If
End Function

Private Function OleInvoke(ByVal pInterface As Long, _
                           ByVal lngCmd As Long, _
                           ParamArray aParam()) As Variant
    If pInterface <> 0& Then
        If UBound(aParam) >= 0& Then
            Dim varParam As Variant
            varParam = aParam
            If IsArray(varParam) Then varParam = varParam(0)
            Dim lngCount As Long
            lngCount = UBound(varParam)
            Dim lngItem As Long
            Dim olePtr(10) As Long
            Dim oleTyp(10) As Integer
            For lngItem = 0& To lngCount
                oleTyp(lngItem) = VarType(varParam(lngItem))
                olePtr(lngItem) = VarPtr(varParam(lngItem))
            Next
        End If
        Dim varRet As Variant
        If DispCallFunc(pInterface, lngCmd * 4, CC_STDCALL, _
                        vbLong, lngItem, VarPtr(oleTyp(0)), _
                        VarPtr(olePtr(0)), varRet) <> S_OK Then
            Debug.Print "Fehler beim Aufrufen der Interface-Funktion!"
        End If

'        If varRet <> S_OK Then
'            Debug.Print "0x" & Hex$(varRet)
'        End If
    End If
    OleInvoke = varRet
End Function

Private Function Str2Guid(ByVal strGUID As String) As GUID
    Call CLSIDFromString(StrPtr(strGUID), Str2Guid)
End Function

Private Function Guid2Str(ByRef tGuid As GUID) As String
    Dim lngRet As Long
    Dim strGUID As String
    strGUID = String(40, 0)
    lngRet = StringFromGUID(tGuid, StrPtr(strGUID), Len(strGUID))
    If lngRet > 0 Then
        strGUID = Left$(strGUID, lngRet - 1)
    Else
        strGUID = vbNullString
    End If
    Guid2Str = strGUID
End Function

Private Function Ptr2Str(ByVal lpStr As Long) As String
    Dim lngLen As Long
    Dim bytBuffer() As Byte
    lngLen = lstrlenW(lpStr) * 2
    If lngLen > 0 Then
        ReDim bytBuffer(lngLen - 1)
        Call CopyMemory(bytBuffer(0), ByVal lpStr, lngLen)
        Call CoTaskMemFree(lpStr)
        Ptr2Str = bytBuffer
    End If
End Function
